!*******************************************************************
!*                         SOURCE                                  *
!*               BOX-ISOTROPIC-GAUSS-SPECTRUM                      *
!*                                                                 *
!* Short description:                                              *
!*   Generation of primary particle states for radiation transport *
!*   calculations with PENELOPE.                                   *
!*******************************************************************


      subroutine BIGSsource(n)
!*******************************************************************
!*    Input:                                                       *
!*      n -> History no.                                           *
!*    Output:                                                      *
!*      The generated particles are pushed to the particle stack.  *
!*      Note that common /track/ is not declared.                  *
!*******************************************************************
      implicit none
      real*8 n

      integer*4 kpar,ibody,mat,ilb
      real*8 e,x,y,z,u,v,w,wght
      common/track/e,x,y,z,u,v,w,wght,kpar,ibody,mat,ilb(5)
      integer*4 ipol
      real*8 sp1,sp2,sp3
      common/stokes/sp1,sp2,sp3,ipol
      integer*4 xvox,yvox,zvox,absvox
      real*8 uold,vold,ivx,ivy,ivz,sdx,sdy,sdz
      common /partvox/ uold,vold,ivx,ivy,ivz,sdx,sdy,sdz,
     &                 xvox,yvox,zvox,absvox
      logical warned,active
      integer parsrc,matsrc,nspc,dim
      parameter (dim=1000)
      integer*4 xipol
      real*8 shots,espc(dim),pspc(dim),despc(dim),esigma
      real*8 rotv(3,3),cossrc,rotbox(3,3),rsrc(3),drsrc(3)
      real*8 xsigma,ysigma,xsp1,xsp2,xsp3
      common /srcbig/ rotbox,rsrc,drsrc,xsigma,ysigma,
     &                rotv,cossrc,espc,pspc,despc,esigma,shots,
     &                xsp1,xsp2,xsp3,xipol,
     &                parsrc,matsrc,nspc,warned,active
      integer ntrial,i,seeki
      integer*4 ncross,bodytmp,mattmp,xvoxtmp,yvoxtmp,zvoxtmp,voxtmp
      real*8 r(3),d(3),phi,dsef,sintheta,radius,offset
      real*8 pi,twopi,one3rd,infty,rand,randno,kfact
      parameter (pi=3.1415926535897932d0,twopi=2.0d0*pi,infty=1.0d30)
      parameter (one3rd=1.0d0/3.0d0,kfact=6.0d0)
      external rand

      if (.not.active) return

      ntrial = 0
      kpar = parsrc  ! Particle type

      ipol = xipol   ! Polarization--effective for photons only
      sp1 = xsp1
      sp2 = xsp2
      sp3 = xsp3

      direction: do
        d(3) = cossrc+(1.0-cossrc)*rand(0.0d0)
        sintheta = sqrt(1.0-d(3)**2)
        phi = twopi*rand(0.0d0)
        d(2) = sintheta*sin(phi)
        d(1) = sintheta*cos(phi)  ! Direction in Sv frame
        d = matmul(rotv,d)        ! Rotate direction vector from Sv to Slab
        u = d(1)                  ! Transfer direction to PENELOPE common
        v = d(2)
        w = d(3)

        position: do
          ntrial = ntrial+1
          if (n.lt.1.01.and.ntrial.gt.100) then
            write(*,*)
     &    'BIGSsource:ERROR: could not generate a valid particle '//
     &    'after 100 trials;'
            write(*,*)
     &       '  box enclosure does not contain the source '//
     &       'material or direction aims at vacuum'
            stop
          endif

          do i=1,3
            r(i) = drsrc(i)*(rand(dble(i))-0.5)    ! Coordinates in Sbox
          enddo
          if (xsigma.ne.0.0d0.or.ysigma.ne.0.0d0) then ! Gaussian x,y displacement
            radius = sqrt(-2.0*log(rand(0.0d0)))   ! Sample using Box-Muller method
            phi = twopi*rand(0.0d0)
            r(1) = r(1)+xsigma*radius*sin(phi)     ! Apply displacement in Sbox frame
            r(2) = r(2)+ysigma*radius*cos(phi)
          endif
          r = matmul(rotbox,r)+rsrc                ! Coordinates in Slab
          x = r(1)                                 ! Transfer position to PENELOPE common
          y = r(2)
          z = r(3)
          call locatex                             ! Finds body and material (needs U,V,W)

          ! Accept or reject?
          if (matsrc.eq.0) then                             ! Any material will do, even vacuum
            if (mat.eq.0) call stepx(infty,dsef,ncross)     ! Vacuum, advance up to object or infinity
            exit direction                                  ! Always accept
          else if (matsrc.eq.mat) then                      ! Materials match
            exit direction                                  ! Accept this point
          else if (matsrc.lt.0) then                        ! Particle must aim at specified material
            if (mat.eq.-matsrc) exit direction              ! Already inside specified material
            bodytmp = ibody                                 ! Store original position variables
            mattmp = mat
            xvoxtmp = xvox
            yvoxtmp = yvox
            zvoxtmp = zvox
            voxtmp  = absvox
            do
              call stepx(infty,dsef,ncross)                 ! Move forward to find specified material
              if (mat.eq.0.or.ncross.eq.0) cycle direction  ! Not aiming at specified material
              if (mat.eq.-matsrc) then                      ! Accept this point
                x = r(1)                                    ! Restore original position
                y = r(2)
                z = r(3)
                ibody  = bodytmp
                mat    = mattmp
                xvox   = xvoxtmp
                yvox   = yvoxtmp
                zvox   = zvoxtmp
                absvox = voxtmp
                if (mat.eq.0) call stepx(infty,dsef,ncross) ! Advance up to object
                exit direction                              ! Accept this point
              endif
            enddo
          endif                                             ! Re-sample XYZ if materials don't match
        enddo position
      enddo direction

      ! Kinetic energy:
      randno = rand(0.0d0)
      i = seeki(pspc,randno,nspc)
      e = espc(i)+(randno-pspc(i))*despc(i)
      if (esigma.ne.0.0) then             ! Gaussian distrib, sample using Box-Muller
        do
          offset =
     &      sqrt(-2.0*log(rand(0.0d0)))*sin(twopi*rand(1.0d0))
          if (abs(offset).lt.kfact) exit  ! Truncate Gauss at +-kfact
        enddo                             ! ...else repeat
        e = max(e+esigma*offset,0.0)      ! Prevent neg values
      endif

      ! Average no. of shots per call:
      shots = shots+ntrial
      if (.not.warned.and.shots.gt.n*10.0) then
        write(*,*) ''
        write(*,'(a)') '***************'
        write(*,'(a)')
     &    'BIGSsource:WARNING: source effectiveness is too low !'
        write(*,'(a)') '  Redefine source parameters appropriately.'
        write(*,'(a)')
     &    '  Number of histories and shots per source call so far:'
        write(*,'(2x,f18.0,1x,es12.5)') n,shots/n
        write(*,'(a)') '***************'
        write(*,*) ''
        warned = .true.
      endif

      wght   = 1.0                              ! Init statistical weight
      ilb(1) = 1                                ! Tag as source particle (i.e. 1st generation)
      ilb(2) = 0                                ! Clear other labels
      ilb(3) = 0
      ilb(4) = 0
      ilb(5) = 0                                ! Optional label (transferred to descendants)

      call stores(e,x,y,z,u,v,w,wght,kpar,ilb,ipol)  ! Push particle to stack
      call tally(0,e)                           ! Deposit its kinetic E
      end


      subroutine BIGSinisrc(activated,emax)
!*******************************************************************
!*    Initializes the source.                                      *
!*                                                                 *
!*    Output:                                                      *
!*      activated -> TRUE if the source is active.                 *
!*      emax -> max source energy (eV)                             *
!*******************************************************************
      implicit none
      logical activated
      real*8 emax

      logical warned,active
      integer parsrc,matsrc,nspc,dim
      parameter (dim=1000)
      integer*4 xipol
      real*8 shots,espc(dim),pspc(dim),despc(dim),esigma
      real*8 rotv(3,3),cossrc,rotbox(3,3),rsrc(3),drsrc(3)
      real*8 xsigma,ysigma,xsp1,xsp2,xsp3
      common /srcbig/ rotbox,rsrc,drsrc,xsigma,ysigma,
     &                rotv,cossrc,espc,pspc,despc,esigma,shots,
     &                xsp1,xsp2,xsp3,xipol,
     &                parsrc,matsrc,nspc,warned,active
      character*80 buffer
      character*(*) secid,eos
      parameter (secid=
     &'[SECTION SOURCE BOX ISOTROPIC GAUSS SPECTRUM v.2009-10-30]')
      parameter (eos='[END OF BIGS SECTION]')
      integer j,error
      real*8 omega,theta,phi,mc2,twomc2,zero,oneeps
      real*8 prob,pi,norm,fwhm2sig,deg2rad,kfact,dsrc(3)
      parameter (pi=3.1415926535897932d0,deg2rad=pi/180.0)
      parameter (fwhm2sig=1.0/(8.0*log(2.0d0))**0.5,kfact=6.0d0)
      parameter (mc2=5.10998918d5,twomc2=2.0*mc2,zero=1.0d-30)
      parameter (oneeps=1.0d0+1.0d-14)

      write(*,*) ' '
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'BIGSinisrc:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      read(*,'(a80)') buffer
      buffer = adjustl(buffer)
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      if (buffer.eq.'ON') then
        active = .true.
        activated = active
      else if (buffer.eq.'OFF') then
        active = .false.
        activated = active
        write(*, '(a)')
     &    '>>>> Source Box Isotropic Gauss Spectrum is OFF >>>>'
        do
          read(*,'(a80)',iostat=error) buffer
          if (error.ne.0) then
            write(*,'(a,a,a)') 'BIGSinisrc:ERROR: ',
     &       'Unable to find End-Of-Section mark: ',eos
            stop
          endif
          if (index(buffer,eos).ne.0) return
        enddo
      else
        write(*,'(a)')
     &    'BIGSinisrc:ERROR: expecting to find ON or OFF'
        write(*,'(a)') 'found instead:'
        write(*,'(a)') buffer
        stop
      endif

      ! Type of particle:
      write(*,'(a)') 'Particle type:'
      read(*,*) parsrc
      write(*,'(1x,i0)') parsrc
      if (parsrc.ne.1.and.parsrc.ne.2.and.parsrc.ne.3) then
        write(*,*) 'BIGSinisrc:ERROR: invalid particle type'
        stop
      endif

      ! Photon polarisation:
      write(*,'(a)') 'Photon polarisation physics:'
      read(*,*) xipol
      if (xipol.ne.1.and.xipol.ne.0) then
        write(*,'(a)')
     &    'BIGSinisrc:ERROR: invalid polarisation switch, '//
     &    'should be 0 or 1'
        stop
      endif
      if (xipol.eq.0) then
        write(*,'(a)') '  not activated'
      else
        write(*,'(a)') '  activated'
      endif
      if (xipol.eq.1) write(*,'(a)') 'Stokes parameters:'
      read(*,*) xsp1,xsp2,xsp3
      if (xipol.eq.1) write(*,'(3(1x,es12.5))') xsp1,xsp2,xsp3
      if (xipol.eq.1.and.xsp1**2+xsp2**2+xsp3**2.gt.oneeps) then
        write(*,*)
     &    'BIGSinisrc:ERROR: vector P={P1,P2,P3} must be P^2 <= 1'
        stop
      endif

      ! Energy:
      write(*,'(a)') 'Reading energy spectrum'
      read(*,'(a80)') buffer
      write(*,'(a)') '  Energy(eV)  Relat.Probability   Bin#'
      nspc = 0
      pspc(1) = 0.0
      do
        nspc = nspc+1
        read(*,*) espc(nspc),prob
        write(*,'(2(1x,es12.5),1x,i5)') espc(nspc),prob,nspc
        if (espc(nspc).lt.0.0) then
          write(*,*) 'BIGSinisrc:ERROR: negative energy'
          stop
        else if (espc(nspc).lt.espc(max(nspc-1,1))) then
          write(*,*) 'BIGSinisrc:ERROR: decreasing energy'
          stop
        endif
        if (prob.lt.0.0) exit    ! End of spectrum
        if (nspc.ge.dim) then
          write(*,*) 'BIGSinisrc:ERROR: too many bins in spectrum;'
          write(*,*) '              enlarge DIM'
          stop
        endif
        pspc(nspc+1) = pspc(nspc)+prob
      enddo
      write(*,'(a)') 'No. of bins read:'
      write(*,'(1x,i0)') nspc-1
      if (nspc.lt.2) then
        write(*,*)
     &    'BIGSinisrc:ERROR: at least 1 bin must be defined'
        stop
      endif
      write(*,'(a)')
     &  'Sum of relative probabilities before normalization:'
      write(*,'(es12.5)') pspc(nspc)
      if (pspc(nspc).le.0.0) then
        write(*,*)
     &    'BIGSinisrc:ERROR: all probabilities are zero.'
        stop
      endif
      do j=1,nspc    ! Normalize to unity
        pspc(j) = pspc(j)/pspc(nspc)
      enddo
      do j=1,nspc-1    ! Prepare auxiliar array for sampling
        despc(j) = 0.0
        if (pspc(j+1).gt.pspc(j))
     &    despc(j) = (espc(j+1)-espc(j))/(pspc(j+1)-pspc(j))
      enddo

      write(*,'(a)') 'FWHM of Gaussian energy distrib. (eV):'
      read(*,*) esigma
      write(*,'(1x,es12.5)') esigma
      esigma = esigma*fwhm2sig
      emax = espc(nspc)+kfact*esigma  ! Set max energy
      if (parsrc.eq.3) emax = emax+twomc2  ! Allowance for e+ annihilation

      ! Position:
      write(*,'(a)') 'Center coordinates (cm):'
      read(*,*) rsrc
      write(*,'(3(1x,es12.5))') rsrc
      write(*,'(a)') 'Box sides (cm):'
      read(*,*) drsrc
      write(*,'(3(1x,es12.5))') drsrc
      write(*,'(a)') 'FWHMs of x and y distribs. (cm):'
      read(*,*) xsigma,ysigma
      write(*,'(2(1x,es12.5))') xsigma,ysigma
      xsigma = xsigma*fwhm2sig
      ysigma = ysigma*fwhm2sig
      write(*,'(a)') 'Box Euler angles (deg):'
      read(*,*) omega,theta,phi
      write(*,'(3(1x,es12.5))') omega,theta,phi
      call rotation(omega*deg2rad,theta*deg2rad,phi*deg2rad,rotbox)  ! Rotation Sbox -> Slab

      ! Material:
      write(*,'(a)') 'Source material:'
      read(*,*) matsrc
      write(*,'(1x,i0)') matsrc
      if (matsrc.lt.0) write(*,'(a)')
     &  '  (taken as the material at which particles must be aiming)'

      ! Direction:
      write(*,'(a)') 'Direction vector (u,v,w):'
      read(*,*) dsrc
      write(*,'(3(1x,es12.5))') dsrc
      norm = sqrt(sum(dsrc**2))
      if (norm.lt.zero) then
        write(*,*) 'BIGSinisrc:ERROR: null direction.'
        stop
      endif
      dsrc = dsrc/norm  ! Direction vector normalized to 1
      call rotation(0.0d0,acos(dsrc(3)),atan2(dsrc(2),dsrc(1)),rotv)  ! Rotation Sv -> Slab
      write(*,'(a)') 'Semi-aperture (deg) and its cos():'
      read(*,*) theta
      cossrc = cos(theta*deg2rad)
      write(*,'(2(1x,es12.5))') theta,cossrc
      if (theta.lt.0.0.or.theta.gt.180.0) then
        write(*,*) 'BIGSinisrc:ERROR: semi-aperture not in [0,180].'
        stop
      endif

      ! Init performance vars:
      shots = 0.0
      warned = .false.

      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,*) 'BIGSinisrc:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') '>>>> BIGS source initialization finished >>>>'
      end


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
