!*******************************************************************
!*                    TALLY ENERGY DEPOSITION                      *
!*                                                                 *
!* Short description:                                              *
!*   Calculates the energy deposited in each material.             *
!*   The so-called collision estimator, which is unambiguously     *
!*   defined in the PENELOPE transport algorithm, is used.         *
!*******************************************************************


      subroutine EDPtally(mode,eloss)
!*******************************************************************
!*    Input:                                                       *
!*      mode -> Identifies the state of the calling procedure      *
!*      eloss -> energy deposition                                 *
!*******************************************************************
      implicit none
      integer mode
      real*8 eloss

      integer*4 kpar,ibody,mat,ilb
      real*8 e,x,y,z,u,v,w,wght
      common/track/e,x,y,z,u,v,w,wght,kpar,ibody,mat,ilb(5)
      logical active
      integer maxmat,detmat
      parameter (maxmat=10)
      real*8 edptmp,edep,edep2,unclimit
      common /scoedp/ edptmp(maxmat),edep(maxmat),edep2(maxmat),
     &                unclimit,detmat,active
      integer i

      if (.not.active) return

      if (mode.le.0.and.mat.gt.0) then        ! There's energy to be deposited
        edptmp(mat) = edptmp(mat)+eloss*wght  ! Update temporary E counter

      else if (mode.eq.6) then        ! End-of-history bookkeeping
        do i=1,maxmat
          if (edptmp(i).gt.0.0) then  ! Transfer temporary counter to mean and variance
            edep(i)   = edep(i) +edptmp(i)
            edep2(i)  = edep2(i)+edptmp(i)**2
            edptmp(i) = 0.0           ! Clear counter to start a fresh history
          endif
        enddo

      endif
      end


      subroutine EDPreport(n,cputim,uncdone)
!*******************************************************************
!*    Input:                                                       *
!*      n -> no. of histories simulated                            *
!*      cputim -> elapsed CPU time                                 *
!*    Output:                                                      *
!*      uncdone -> 2 if uncert reached, 1 if not defined, 0 else   *
!*******************************************************************
      implicit none
      integer uncdone
      real*8 n,cputim

      integer*4 seed1,seed2
      common/rseed/seed1,seed2
      logical active
      integer maxmat,detmat
      parameter (maxmat=10)
      real*8 edptmp,edep,edep2,unclimit
      common /scoedp/ edptmp(maxmat),edep(maxmat),edep2(maxmat),
     &                unclimit,detmat,active
      integer i,out,finduf,error
      real*8 q,q2,sigma,eff,uncert,invn

      uncdone = 1
      if (.not.active) return

      call EDPdump(1)  ! Sim dump file

      ! Prepare output files:
      out = finduf()
      open(out,file='tallyEnergyDeposition.dat',iostat=error)
      if (error.ne.0) then
        write(*,*)
        write(*,'(a)')
     &    '*********************************************'
        write(*,'(a)')
     &    'EDPreport:ERROR: cannot open output data file'
        write(*,'(a)')
     &    '*********************************************'
        close(out)  ! Just in case
        return
      endif

      write(out,'(a)')
     &'#>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      write(out,'(a)') '# [SECTION REPORT ENERGY DEPOSITION]'
      write(out,'(a)')
     &   '# Material : Total Energy Deposition(eV) : +-2sigma'

      invn = 1.0/n
      do i=1,maxmat
        q  = edep(i)*invn
        q2 = edep2(i)*invn
        sigma = (q2-q**2)*invn
        sigma = sqrt(max(sigma,0.0))
        write(out,'(1x,i3,1x,es12.5,1x,es8.1)') i,q,2.0*sigma
      enddo

      ! Evaluate rel. uncertainty:
      uncert = 200.0
      q  = edep(detmat)*invn
      q2 = edep2(detmat)*invn
      sigma = (q2-q**2)*invn
      sigma = sqrt(max(sigma,0.0))
      if (q.gt.0.0) uncert = 200.0*sigma/q

      ! Generic report:
      write(out,'(a)') ' '
      write(out,'(a)') '# Performance report'
      write(out,'(a)') '#   Random seeds:'
      write(out,'(a,i10)') '#   ',seed1
      write(out,'(a,i10)') '#   ',seed2
      write(out,'(a)') '#   No. of histories simulated [N]:'
      write(out,'(a,f18.0)') '#   ',n
      write(out,'(a)') '#   CPU time [t] (s):'
      write(out,'(a,es12.5)') '#   ',cputim
      if (cputim.gt.0.0) then
        write(out,'(a)') '#   Speed (histories/s):'
        write(out,'(a,es12.5)') '#   ',n/cputim
      endif
      write(out,'(a)')
     & '#   Uncertainty of energy deposition in detector material, '//
     & 'in % [uncert]:'
      write(out,'(a,es12.5)') '#   ',uncert
      eff = n*uncert**2
      if (eff.gt.0.0) then
        write(out,'(a)') '#   Intrinsic efficiency [N*uncert^2]^-1:'
        write(out,'(a,es12.5)') '#   ',1.0/eff
      endif
      eff = cputim*uncert**2
      if (eff.gt.0.0) then
        write(out,'(a)') '#   Absolute efficiency [t*uncert^2]^-1:'
        write(out,'(a,es12.5)') '#   ',1.0/eff
      endif
      write(out,'(a)') '#'
      write(out,'(a)') '# Have a nice day.'
      close(out)
      end


      subroutine EDPinitally
!*******************************************************************
!*    Initializes. To be called before TALLY.                      *
!*******************************************************************
      implicit none
      logical active
      integer maxmat,detmat
      parameter (maxmat=10)
      real*8 edptmp,edep,edep2,unclimit
      common /scoedp/ edptmp(maxmat),edep(maxmat),edep2(maxmat),
     &                unclimit,detmat,active
      character*(*) secid,eos
      parameter (secid=
     &'[SECTION TALLY ENERGY DEPOSITION v.2012-06-01]')
      parameter (eos='[END OF EDP SECTION]')
      character*80 buffer
      integer error

      write(*,*) ' '
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'EDPinitally:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      read(*,'(a80)') buffer
      buffer = adjustl(buffer)
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      if (buffer.eq.'ON') then
        active = .true.
      else if (buffer.eq.'OFF') then
        active = .false.
        write(*, '(a)') '>>>> Tally Energy Deposition is OFF'
        do
          read(*,'(a80)',iostat=error) buffer
          if (error.ne.0) then
            write(*,'(a,a,a)') 'EDPinitally:ERROR: ',
     &       'Unable to find End-Of-Section mark: ',eos
            stop
          endif
          if (index(buffer,eos).ne.0) return
        enddo
      else
        write(*,'(a)')
     &    'EDPinitally:ERROR: expecting to find ON or OFF'
        write(*,'(a)') 'found instead:'
        write(*,'(a)') buffer
        stop
      endif

      write(*,'(a)') 'Detection material set to:'
      read(*,*) detmat
      write(*,'(i3)') detmat

      write(*,'(a)') 'Relative uncertainty (%) requested:'
      read(*,*) unclimit
      write(*,'(1x,es12.5)') unclimit

      ! Clear counters:
      edptmp = 0.0
      edep   = 0.0
      edep2  = 0.0

      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,*) 'EDPinitally:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif

      call EDPdump(0)  ! Sim restart file

      write(*,'(a)') '>>>> EDP tally initialization finished >>>>'
      end


      subroutine EDPdump(mode)
!*******************************************************************
!*    Dumps into or reads data from a dump file.                   *
!*                                                                 *
!*    Input:                                                       *
!*      mode -> 1 to write dump file, else to read from it.        *
!*******************************************************************
      implicit none
      integer mode

      logical active
      integer maxmat,detmat
      parameter (maxmat=10)
      real*8 edptmp,edep,edep2,unclimit
      common /scoedp/ edptmp(maxmat),edep(maxmat),edep2(maxmat),
     &                unclimit,detmat,active
      character*80 dumpfilen
      integer restartfile,dumpfile
      common /dump/ restartfile,dumpfile,dumpfilen

      if (mode.eq.1) then
        if (dumpfile.lt.0) return  ! No dump file open
        write(dumpfile) edep(1:maxmat),edep2(1:maxmat)
      else
        if (restartfile.lt.0) return  ! No restart file open
        read(restartfile) edep(1:maxmat),edep2(1:maxmat)
      endif
      end


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
