!*******************************************************************
!*                   TALLY PULSE HEIGHT SPECTRUM                   *
!*                                                                 *
!* Short description:                                              *
!*   Calculates the spectrum of the energy deposited in a given    *
!*   material.                                                     *
!*******************************************************************


      subroutine PHStally(mode,eloss)
!*******************************************************************
!*    Input:                                                       *
!*      mode -> Identifies the state of the calling procedure      *
!*      eloss -> energy deposition                                 *
!*******************************************************************
      implicit none
      integer mode
      real*8 eloss

      integer*4 kpar,ibody,mat,ilb
      real*8 e,x,y,z,u,v,w,wght
      common/track/e,x,y,z,u,v,w,wght,kpar,ibody,mat,ilb(5)
      logical active,isGaussConvo
      integer detmat,maxbin,nbin
      parameter (maxbin=1000)
      real*8 edptmp,count,count2,emin,ebin,iebin,unclimit,a0,a1
      common /scophs/ count(maxbin),count2(maxbin),edptmp,emin,ebin,
     &                iebin,a0,a1,unclimit,nbin,detmat,isGaussConvo,
     &                active
      integer bin
      real*8 pi,dospi,sigma,rand,econvol
      parameter (pi=3.1415926535897932d0,dospi=2.0*pi)
      external rand

      if (.not.active) return

      if (mode.le.0.and.mat.eq.detmat) then
        edptmp = edptmp+eloss            ! Update temporary E counter

      else if (mode.eq.6) then           ! End-of-history bookkeeping

        if (isGaussConvo) then           ! Sample deposited energy from a Gaussian
          sigma = sqrt(a0+a1*edptmp)
          econvol = edptmp+sigma*        ! Box-Muller method to sample a Gaussian
     &           sin(dospi*rand(0.0d0))*sqrt(-2.0*log(rand(1.0d0)))
          bin = (econvol-emin)*iebin+1.0 ! E bin number
        else
          bin =  (edptmp-emin)*iebin+1.0 ! Straight bin assigment, no Gaussian
        endif
        if (bin.gt.0.and.bin.le.nbin) then
          count(bin)  = count(bin) +wght ! Score one count
          count2(bin) = count2(bin)+wght**2  ! Score variance
        endif
        edptmp = 0.0                     ! Clear counter to start a fresh history

      endif
      end


      subroutine PHSreport(n,cputim,uncdone)
!*******************************************************************
!*    Input:                                                       *
!*      n -> no. of histories simulated                            *
!*      cputim -> elapsed CPU time                                 *
!*    Output:                                                      *
!*      uncdone -> 2 if uncert reached, 1 if not defined, 0 else   *
!*******************************************************************
      implicit none
      integer uncdone
      real*8 n,cputim

      integer*4 seed1,seed2
      common/rseed/seed1,seed2
      logical active,isGaussConvo
      integer detmat,maxbin,nbin
      parameter (maxbin=1000)
      real*8 edptmp,count,count2,emin,ebin,iebin,unclimit,a0,a1
      common /scophs/ count(maxbin),count2(maxbin),edptmp,emin,ebin,
     &                iebin,a0,a1,unclimit,nbin,detmat,isGaussConvo,
     &                active
      integer i,nchan,out,finduf,error
      real*8 q,q2,sigma,eff,avesig,maxq,emiddle,elow,uncert,invn

      uncdone = 1
      if (.not.active) return

      call PHSdump(1)  ! Sim dump file

      ! Prepare output files:
      out = finduf()
      open(out,file='tallyPulseHeightSpectrum.dat',iostat=error)
      if (error.ne.0) then
        write(*,*)
        write(*,'(a)')
     &    '*********************************************'
        write(*,'(a)')
     &    'PHSreport:ERROR: cannot open output data file'
        write(*,'(a)')
     &    '*********************************************'
        close(out)  ! Just in case
        return
      endif

      ! Evaluate 1/2 of the max score:
      avesig = 0.0
      nchan = 0
      maxq = 0.0
      do i=1,nbin
        maxq = max(maxq,count(i))
      enddo
      maxq = 0.5*maxq

      write(out,'(a)')
     &'#>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      write(out,'(a)') '# [SECTION REPORT PULSE HEIGHT SPECTRUM]'
      write(out,'(a)') '#'
      write(out,'(a)')
     &  '# For plotting purposes, two energies per bin are given,'
      write(out,'(a)')
     &  '#   namely, the low end and the middle point of each bin'
      write(out,'(a)') '#'
      write(out,'(a)')
     &  '# Elow(eV) : Emiddle(eV) : counts(1/eV) : +-2sigma'
      invn = 1.0/n
      do i=1,nbin
        q = count(i)*invn
        q2 = count2(i)*invn
        sigma = (q2-q**2)*invn
        sigma = sqrt(max(sigma,0.0))
        q = q*iebin
        sigma = sigma*iebin
        elow = emin+ebin*(i-1)
        emiddle = elow+ebin*0.5
        write(out,'(3(1x,es12.5),1x,es8.1)')
     &    elow,emiddle,q,2.0*sigma
        ! Evaluate average uncertainty for scores above 1/2 max score:
        if (count(i).gt.maxq) then
          avesig = avesig+(sigma/q)**2
          nchan = nchan+1
        endif
      enddo
      uncdone = 0
      if (nchan.gt.0) then
        uncert = 200.0*sqrt(avesig/nchan)
        if (uncert.lt.unclimit) uncdone = 2  ! Uncertainty reached
      else
        uncert = 0.0  ! Uncertainty assumed not reached when score is nil
      endif

      ! Generic report:
      write(out,'(a)') ' '
      write(out,'(a)') '# Performance report'
      write(out,'(a)') '#   Random seeds:'
      write(out,'(a,i10)') '#   ',seed1
      write(out,'(a,i10)') '#   ',seed2
      write(out,'(a)') '#   No. of histories simulated [N]:'
      write(out,'(a,f18.0)') '#   ',n
      write(out,'(a)') '#   CPU time [t] (s):'
      write(out,'(a,es12.5)') '#   ',cputim
      if (cputim.gt.0.0) then
        write(out,'(a)') '#   Speed (histories/s):'
        write(out,'(a,es12.5)') '#   ',n/cputim
      endif
      write(out,'(a)')
     & '#   Average uncertainty (above 1/2 max score) in % [uncert]:'
      write(out,'(a,es12.5)') '#   ',uncert
      eff = n*uncert**2
      if (eff.gt.0.0) then
        write(out,'(a)') '#   Intrinsic efficiency [N*uncert^2]^-1:'
        write(out,'(a,es12.5)') '#   ',1.0/eff
      endif
      eff = cputim*uncert**2
      if (eff.gt.0.0) then
        write(out,'(a)') '#   Absolute efficiency [t*uncert^2]^-1:'
        write(out,'(a,es12.5)') '#   ',1.0/eff
      endif
      write(out,'(a)') '#'
      write(out,'(a)') '# Have a nice day.'
      close(out)
      end


      subroutine PHSinitally
!*******************************************************************
!*    Initializes. To be called before TALLY.                      *
!*******************************************************************
      implicit none
      logical active,isGaussConvo
      integer detmat,maxbin,nbin
      parameter (maxbin=1000)
      real*8 edptmp,count,count2,emin,ebin,iebin,unclimit,a0,a1
      common /scophs/ count(maxbin),count2(maxbin),edptmp,emin,ebin,
     &                iebin,a0,a1,unclimit,nbin,detmat,isGaussConvo,
     &                active
      character*(*) secid,eos
      parameter (secid=
     &'[SECTION TALLY PULSE HEIGHT SPECTRUM v.2012-06-01]')
      parameter (eos='[END OF PHS SECTION]')
      character*80 buffer
      integer error
      real*8 emax,fwhm2sig2
      parameter (fwhm2sig2=1.0/(8.0*log(2.0d0)))

      write(*,*) ' '
      write(*,'(a)')
     & '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>'
      call getline(buffer)
      if (index(buffer,secid).eq.0) then
        write(*,'(a)') 'PHSinitally:ERROR: incorrect section header;'
        write(*,'(a,a)') '  expecting to find: ',secid
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif
      write(*,'(a)') secid

      read(*,'(a80)') buffer
      buffer = adjustl(buffer)
      buffer = buffer(1:scan(buffer,' ')) ! Clip at 1st blank
      if (buffer.eq.'ON') then
        active = .true.
      else if (buffer.eq.'OFF') then
        active = .false.
        write(*, '(a)') '>>>> Tally Pulse Height Spectrum is OFF'
        do
          read(*,'(a80)',iostat=error) buffer
          if (error.ne.0) then
            write(*,'(a,a,a)') 'PHSinitally:ERROR: ',
     &       'Unable to find End-Of-Section mark: ',eos
            stop
          endif
          if (index(buffer,eos).ne.0) return
        enddo
      else
        write(*,'(a)')
     &    'PHSinitally:ERROR: expecting to find ON or OFF'
        write(*,'(a)') 'found instead:'
        write(*,'(a)') buffer
        stop
      endif

      write(*,'(a)') 'Detection material set to:'
      read(*,*) detmat
      write(*,'(i3)') detmat

      write(*,'(a)')
     &  'Spectrum limits [Emin,Emax] (eV) and no. bins:'
      read(*,*) emin,emax,nbin
      write(*,'(2(1x,es12.5),1x,i5)') emin,emax,nbin
      ebin = (emax-emin)/nbin
      iebin = 1.0/ebin
      ! Add one bin to store, separately, counts with E=Emax:
      nbin = nbin+1
      if (nbin.gt.maxbin) then
        write(*,*)
     &    'PHSinitally:ERROR: Too many bins; increase MAXBIN'
        stop
      endif

      write(*,'(a)') 'Parameters for Gaussian convolution:'
      read(*,*) a0,a1
      write(*,'(2(1x,es12.5))') a0,a1
      if (min(a0,a1).lt.0.0) then
        write(*,*)
     &    'PHSinitally:ERROR: Gaussian parameters cannot be negative.'
        stop
      endif
      isGaussConvo = .false.
      if (max(a0,a1).gt.0.0) isGaussConvo = .true.
      a0 = a0*fwhm2sig2  ! This is in units of sigma^2
      a1 = a1*fwhm2sig2

      write(*,'(a)') 'Relative uncertainty (%) requested:'
      read(*,*) unclimit
      write(*,'(1x,es12.5)') unclimit

      ! Clear counters:
      edptmp = 0.0
      count  = 0.0
      count2 = 0.0

      read(*,'(a80)') buffer
      if (index(buffer,eos).eq.0) then
        write(*,*) 'PHSinitally:ERROR: End-Of-Section mark not found'
        write(*,'(a,a)') '  expecting to find: ',eos
        write(*,'(a,a)') '  found instead:     ',buffer
        stop
      endif

      call PHSdump(0)  ! Sim restart file

      write(*,'(a)') '>>>> PHS tally initialization finished >>>>'
      end


      subroutine PHSdump(mode)
!*******************************************************************
!*    Dumps into or reads data from a dump file.                   *
!*                                                                 *
!*    Input:                                                       *
!*      mode -> 1 to write dump file, else to read from it.        *
!*******************************************************************
      implicit none
      integer mode

      logical active,isGaussConvo
      integer detmat,maxbin,nbin
      parameter (maxbin=1000)
      real*8 edptmp,count,count2,emin,ebin,iebin,unclimit,a0,a1
      common /scophs/ count(maxbin),count2(maxbin),edptmp,emin,ebin,
     &                iebin,a0,a1,unclimit,nbin,detmat,isGaussConvo,
     &                active
      character*80 dumpfilen
      integer restartfile,dumpfile
      common /dump/ restartfile,dumpfile,dumpfilen

      if (mode.eq.1) then
        if (dumpfile.lt.0) return  ! No dump file open
        write(dumpfile) count(1:nbin),count2(1:nbin)
      else
        if (restartfile.lt.0) return  ! No restart file open
        read(restartfile) count(1:nbin),count2(1:nbin)
      endif
      end


!>>>> End Of File >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
